'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var ts = require('typescript');

function bundleComponentModeStyles(config, mode) {
    return Promise.all(mode.styleUrls.map(function (styleUrl) {
        return bundleComponentModeStyle(config, styleUrl);
    })).then(function (results) {
        return mode.styles = results.join('');
    });
}
function bundleComponentModeStyle(config, styleUrl) {
    return new Promise(function (resolve, reject) {
        var scssFilePath = config.packages.path.join(config.srcDir, styleUrl);
        var sassConfig = {
            file: scssFilePath,
            outputStyle: 'compressed'
        };
        if (config.debug) {
            console.log("bundle, render sass: " + scssFilePath);
        }
        config.packages.nodeSass.render(sassConfig, function (err, result) {
            if (err) {
                reject(err);
            }
            else {
                var css = result.css.toString().replace(/\n/g, '').trim();
                resolve(css);
            }
        });
    });
}

function getBundleFileName(bundleId) {
    return "ionic." + bundleId + ".js";
}
function getBundleContent(bundleId, componentModeLoader) {
    return "Ionic.loadComponents(\n/** bundleId **/\n'" + bundleId + "'," + componentModeLoader + ");";
}
function getComponentModeLoader(component, mode) {
    var tag = component.tag.trim().toLowerCase();
    var modeName = (mode.name ? mode.name.trim().toLowerCase() : '');
    var styles = (mode.styles ? mode.styles.replace(/'/g, '"') : '');
    var componentClass = component.componentClass;
    var componentFn = component.componentImporter.trim();
    var watches = JSON.stringify(component.watches);
    var label = tag;
    if (mode.name) {
        label += '.' + mode.name;
    }
    var t = [
        "/** " + label + ": tagName [0] **/\n'" + tag + "'",
        "/** " + label + ": component class name [1] **/\n'" + componentClass + "'",
        "/** " + label + ": watches [2] **/\n" + watches,
        "/** " + label + ": modeName [3] **/\n'" + modeName + "'",
        "/** " + label + ": styles [4] **/\n'" + styles + "'",
        "/** " + label + ": importComponent function [5] **/\n" + componentFn
    ];
    return "\n\n/***************** " + label + " *****************/\n[\n" + t.join(',\n\n') + "\n\n]";
}
function getRegistryContent(registry) {
    var content = '(window.Ionic = window.Ionic || {}).components = ';
    content += JSON.stringify(registry, null, 2) + ';';
    return content;
}
function getBundleId(bundleIndex) {
    return bundleIndex.toString();
}

function getFileMeta(config, ctx, filePath) {
    var fileMeta = ctx.files.get(filePath);
    if (fileMeta) {
        return Promise.resolve(fileMeta);
    }
    return readFile(config.packages, filePath).then(function (srcText) {
        return createFileMeta(config.packages, ctx, filePath, srcText);
    });
}
function createFileMeta(packages, ctx, filePath, srcText) {
    var fileMeta = {
        fileName: packages.path.basename(filePath),
        filePath: filePath,
        fileExt: packages.path.extname(filePath),
        srcDir: packages.path.dirname(filePath),
        srcText: srcText,
        jsFilePath: null,
        jsText: null,
        isTsSourceFile: isTsSourceFile(filePath),
        hasCmpClass: false,
        cmpMeta: null,
        cmpClassName: null
    };
    if (fileMeta.isTsSourceFile) {
        fileMeta.hasCmpClass = hasCmpClass(fileMeta.srcText);
    }
    ctx.files.set(filePath, fileMeta);
    return fileMeta;
}
function readFile(packages, filePath) {
    return new Promise(function (resolve, reject) {
        packages.fs.readFile(filePath, 'utf-8', function (err, data) {
            if (err) {
                reject(err);
            }
            else {
                resolve(data);
            }
        });
    });
}
function writeFile(packages, filePath, content) {
    return new Promise(function (resolve, reject) {
        packages.fs.writeFile(filePath, content, function (err) {
            if (err) {
                reject(err);
            }
            else {
                resolve();
            }
        });
    });
}

function writeFiles(packages, files) {
    var paths = [];
    files.forEach(function (content, filePath) {
        content;
        paths.push(filePath);
    });
    return ensureDirs(packages, paths).then(function () {
        var promises = [];
        files.forEach(function (content, filePath) {
            promises.push(writeFile(packages, filePath, content));
        });
        return Promise.all(promises);
    });
}
function ensureDirs(packages, paths) {
    var path = packages.path;
    var fs = packages.fs;
    var checkDirs = [];
    paths.forEach(function (p) {
        var dir = path.dirname(p);
        if (checkDirs.indexOf(dir) === -1) {
            checkDirs.push(dir);
        }
    });
    checkDirs = checkDirs.sort(function (a, b) {
        if (a.split(path.sep).length < b.split(path.sep).length) {
            return -1;
        }
        if (a.split(path.sep).length > b.split(path.sep).length) {
            return 1;
        }
        if (a.length < b.length) {
            return -1;
        }
        if (a.length > b.length) {
            return 1;
        }
        if (a < b) {
            return -1;
        }
        if (a > b) {
            return 1;
        }
        return 0;
    });
    var dirExists = new Set();
    return new Promise(function (resolve, reject) {
        function checkDir(resolve) {
            var dir = checkDirs.shift();
            if (!dir) {
                resolve();
                return;
            }
            var chunks = dir.split(path.sep);
            checkChunk(chunks, 0, resolve);
        }
        function checkChunk(chunks, appendIndex, resolve) {
            if (appendIndex >= chunks.length - 1) {
                checkDir(resolve);
                return;
            }
            var dir = chunks.slice(0, appendIndex + 2).join(path.sep);
            if (dirExists.has(dir)) {
                checkChunk(chunks, ++appendIndex, resolve);
                return;
            }
            fs.access(dir, function (err) {
                if (err) {
                    fs.mkdir(dir, function (err) {
                        if (err) {
                            reject(err);
                        }
                        else {
                            checkChunk(chunks, ++appendIndex, resolve);
                        }
                    });
                }
                else {
                    dirExists.add(dir);
                    checkChunk(chunks, ++appendIndex, resolve);
                }
            });
        }
        checkDir(resolve);
    });
}
function isTsSourceFile(filePath) {
    var parts = filePath.toLowerCase().split('.');
    if (parts.length > 1) {
        if (parts[parts.length - 1] === 'ts') {
            if (parts.length > 2 && parts[parts.length - 2] === 'd') {
                return false;
            }
            return true;
        }
    }
    return false;
}
function hasCmpClass(sourceText) {
    return (sourceText.indexOf('@Component') > -1);
}
function logError(results, msg) {
    results.errors = results.errors || [];
    results.errors.push(msg);
    return results;
}

function bundle(config, ctx) {
    if (ctx === void 0) { ctx = {}; }
    if (!config.packages) {
        throw 'config.packages required';
    }
    if (!config.packages.fs) {
        throw 'config.packages.fs required';
    }
    if (!config.packages.path) {
        throw 'config.packages.path required';
    }
    if (!config.packages.nodeSass) {
        throw 'config.packages.nodeSass required';
    }
    if (!config.packages.rollup) {
        throw 'config.packages.rollup required';
    }
    if (!config.packages.typescript) {
        throw 'config.packages.typescript required';
    }
    if (config.debug) {
        console.log("bundle, srcDir: " + config.srcDir);
        console.log("bundle, destDir: " + config.destDir);
    }
    ctx.results = {};
    return getManifest(config, ctx).then(function (manifest) {
        var components = getComponents(ctx, manifest);
        return Promise.all(Object.keys(components).map(function (tag) {
            return bundleComponent(config, components[tag]);
        }))
            .then(function () {
            return buildCoreJs(config, ctx, manifest);
        }).then(function () {
            return ctx.results;
        });
    });
}
function bundleComponent(config, component) {
    return bundleComponentModule(config, component).then(function () {
        var modeNames = Object.keys(component.modes);
        return Promise.all(modeNames.map(function (modeName) {
            component.modes[modeName].name = modeName;
            return bundleComponentMode(config, component, component.modes[modeName]);
        }));
    });
}
function bundleComponentModule(config, component) {
    if (component.componentImporter) {
        return Promise.resolve();
    }
    var entry = config.packages.path.join(config.srcDir, component.componentUrl);
    if (config.debug) {
        console.log("bundle, bundleComponentModule, entry: " + entry);
    }
    return config.packages.rollup.rollup({
        entry: entry
    }).then(function (bundle) {
        var results = bundle.generate({
            format: 'iife',
            moduleName: 'ionicModule'
        });
        var code = results.code.trim();
        code = code.replace('(function (exports) {', '');
        code = code.replace('}((this.ionicModule = this.ionicModule || {})));', '');
        code = code.trim();
        code = "function importComponent(exports, h, Ionic) {\n" + code + "\n}";
        component.componentImporter = code;
    });
}
function bundleComponentMode(config, component, mode) {
    if (config.debug) {
        console.log("bundle, bundleComponentMode: " + component.tag + ", " + mode.name);
    }
    return bundleComponentModeStyles(config, mode).then(function () {
        return getComponentModeLoader(component, mode);
    });
}
function getComponents(ctx, manifest) {
    if (!ctx.components) {
        ctx.components = {};
        Object.keys(manifest.components).forEach(function (tag) {
            ctx.components[tag] = Object.assign({}, manifest.components[tag]);
            ctx.components[tag].tag = tag;
        });
    }
    return ctx.components;
}
function buildCoreJs(config, ctx, manifest) {
    ctx.bundles = [];
    manifest.bundles.forEach(function (bundleComponentTags) {
        buildComponentBundles(ctx, bundleComponentTags);
    });
    return generateBundleFiles(config, ctx).then(function () {
        var content = getRegistryContent(ctx.registry);
        var promises = [];
        Object.keys(ctx.manifest.coreFiles).forEach(function (coreDirName) {
            var corePath = ctx.manifest.coreFiles[coreDirName];
            promises.push(createCoreJs(config, content, corePath));
        });
        return Promise.all(promises);
    });
}
function buildComponentBundles(ctx, bundleComponentTags) {
    var allModeNames = getAllModeNames(ctx);
    allModeNames.forEach(function (modeName) {
        var bundle = {
            components: []
        };
        bundleComponentTags.forEach(function (bundleComponentTag) {
            var component = ctx.components[bundleComponentTag];
            if (!component)
                return;
            var mode = component.modes[modeName];
            if (!mode)
                return;
            bundle.components.push({
                component: component,
                mode: mode
            });
        });
        if (bundle.components.length) {
            ctx.bundles.push(bundle);
        }
    });
    bundleComponentTags.forEach(function (bundleComponentTag) {
        var bundle = {
            components: []
        };
        var component = ctx.components[bundleComponentTag];
        if (!component)
            return;
        if (Object.keys(component.modes).length > 1)
            return;
        bundle.components.push({
            component: component,
            mode: {
                name: ''
            }
        });
        if (bundle.components.length) {
            ctx.bundles.push(bundle);
        }
    });
}
function generateBundleFiles(config, ctx) {
    ctx.registry = {};
    return Promise.all(ctx.bundles.map(function (bundle, bundleIndex) {
        var componentModeLoaders = bundle.components.map(function (bundleComponent) {
            return getComponentModeLoader(bundleComponent.component, bundleComponent.mode);
        }).join(',\n');
        bundle.id = getBundleId(bundleIndex);
        bundle.fileName = getBundleFileName(bundle.id);
        bundle.filePath = config.packages.path.join(config.destDir, bundle.fileName);
        bundle.content = getBundleContent(bundle.id, componentModeLoaders);
        bundle.components.forEach(function (bundleComponent) {
            var tag = bundleComponent.component.tag;
            var modeName = bundleComponent.mode.name;
            ctx.registry[tag] = ctx.registry[tag] || [];
            var modes = ctx.registry[tag][0] || {};
            modes[modeName] = bundle.id;
            ctx.registry[tag][0] = modes;
            if (ctx.registry[tag].length === 1 && Object.keys(bundleComponent.component.props).length) {
                ctx.registry[tag].push(bundleComponent.component.props);
            }
        });
        var content = bundle.content;
        if (!config.devMode) {
            try {
                var minifyResults = config.packages.uglify.minify(content, {
                    fromString: true
                });
                content = minifyResults.code;
            }
            catch (e) {
                console.log("uglify.minify error: " + e);
            }
        }
        return writeFile(config.packages, bundle.filePath, content);
    }));
}
function createCoreJs(config, registryContent, srcFilePath) {
    var fileName = config.packages.path.basename(srcFilePath);
    if (config.devMode) {
        srcFilePath = srcFilePath.replace('.js', '.dev.js');
    }
    srcFilePath = config.packages.path.join(config.srcDir, srcFilePath);
    var destFilePath = config.packages.path.join(config.destDir, fileName);
    return readFile(config.packages, srcFilePath).then(function (coreJsContent) {
        var content;
        if (config.devMode) {
            content = registryContent + '\n\n' + coreJsContent;
        }
        else {
            registryContent = registryContent.replace(/\s/g, '');
            content = registryContent + '\n' + coreJsContent;
        }
        if (config.debug) {
            console.log("bundle, createCoreJs: " + destFilePath);
        }
        return writeFile(config.packages, destFilePath, content);
    });
}
function getAllModeNames(ctx) {
    var allModeNames = [];
    Object.keys(ctx.components).forEach(function (tag) {
        var component = ctx.components[tag];
        Object.keys(component.modes).forEach(function (modeName) {
            if (allModeNames.indexOf(modeName) === -1) {
                allModeNames.push(modeName);
            }
        });
    });
    return allModeNames;
}
function getManifest(config, ctx) {
    if (ctx.manifest) {
        return Promise.resolve(ctx.manifest);
    }
    var manifestFilePath = config.packages.path.join(config.srcDir, 'manifest.json');
    if (config.debug) {
        console.log("bundle, manifestFilePath: " + manifestFilePath);
    }
    return readFile(config.packages, manifestFilePath).then(function (manifestStr) {
        return ctx.manifest = JSON.parse(manifestStr);
    });
}

function componentClass(ctx) {
    return function (transformContext) {
        function visitClass(fileMeta, classNode) {
            fileMeta.cmpMeta = getComponentDecoratorData(classNode);
            if (fileMeta.cmpMeta) {
                fileMeta.hasCmpClass = true;
                fileMeta.cmpClassName = classNode.name.getText().trim();
                getPropertyDecoratorMeta(fileMeta.cmpMeta, classNode);
                getWatchDecoratorMeta(fileMeta.cmpMeta, classNode);
                return removeClassDecorator(classNode);
            }
            fileMeta.hasCmpClass = false;
            return classNode;
        }
        function removeClassDecorator(classNode) {
            var classWithoutDecorators = ts.createClassDeclaration(undefined, classNode.modifiers, classNode.name, classNode.typeParameters, classNode.heritageClauses, classNode.members);
            return classWithoutDecorators;
        }
        function getPropertyDecoratorMeta(cmpMeta, classNode) {
            var propMembers = classNode.members.filter(function (n) { return n.decorators && n.decorators.length; });
            cmpMeta.props = cmpMeta.props || {};
            propMembers.forEach(function (memberNode) {
                var isProp = false;
                var propName = null;
                var type = null;
                memberNode.forEachChild(function (n) {
                    if (n.kind === ts.SyntaxKind.Decorator && n.getChildCount() > 1 && n.getChildAt(1).getFirstToken().getText() === 'Prop') {
                        isProp = true;
                    }
                    else if (isProp) {
                        if (n.kind === ts.SyntaxKind.Identifier) {
                            propName = n.getText();
                        }
                        else if (n.kind === ts.SyntaxKind.BooleanKeyword) {
                            type = 'boolean';
                        }
                        else if (n.kind === ts.SyntaxKind.NumberKeyword) {
                            type = 'number';
                        }
                    }
                });
                if (isProp && propName) {
                    cmpMeta.props[propName] = {};
                    if (type) {
                        cmpMeta.props[propName].type = type;
                    }
                    memberNode.decorators = undefined;
                }
            });
        }
        function getWatchDecoratorMeta(cmpMeta, classNode) {
            var propMembers = classNode.members.filter(function (n) { return n.decorators && n.decorators.length; });
            cmpMeta.watches = cmpMeta.watches || {};
            propMembers.forEach(function (memberNode) {
                var isWatch = false;
                var propName = null;
                var methodName = null;
                memberNode.forEachChild(function (n) {
                    if (n.kind === ts.SyntaxKind.Decorator && n.getChildCount() > 1 && n.getChildAt(1).getFirstToken().getText() === 'Watch') {
                        isWatch = true;
                        n.getChildAt(1).forEachChild(function (n) {
                            if (n.kind === ts.SyntaxKind.StringLiteral) {
                                propName = n.getText();
                                propName = propName.replace(/\'/g, '');
                                propName = propName.replace(/\"/g, '');
                                propName = propName.replace(/\`/g, '');
                            }
                        });
                    }
                    else if (isWatch) {
                        if (n.kind === ts.SyntaxKind.Identifier) {
                            methodName = n.getText();
                        }
                    }
                });
                if (isWatch && propName && methodName) {
                    cmpMeta.watches[propName] = {
                        fn: methodName
                    };
                    memberNode.decorators = undefined;
                }
            });
        }
        function visit(fileMeta, node) {
            switch (node.kind) {
                case ts.SyntaxKind.ClassDeclaration:
                    return visitClass(fileMeta, node);
                default:
                    return ts.visitEachChild(node, function (node) {
                        return visit(fileMeta, node);
                    }, transformContext);
            }
        }
        return function (tsSourceFile) {
            var fileMeta = ctx.files.get(tsSourceFile.fileName);
            if (fileMeta && fileMeta.hasCmpClass) {
                return visit(fileMeta, tsSourceFile);
            }
            return tsSourceFile;
        };
    };
}
function getComponentDecoratorData(classNode) {
    var metaData = null;
    if (!classNode.decorators) {
        return metaData;
    }
    var isComponent = false;
    classNode.decorators.forEach(function (decorator) {
        decorator.forEachChild(function (decoratorChild) {
            decoratorChild.forEachChild(function (componentChild) {
                if (componentChild.getText().trim() === 'Component') {
                    isComponent = true;
                }
                else if (isComponent) {
                    metaData = parseComponentMetaData(componentChild.getText());
                }
            });
        });
    });
    return metaData;
}
function parseComponentMetaData(text) {
    try {
        var fnStr = "return " + text + ";";
        var cmpMeta = new Function(fnStr)();
        return updateComponentMeta(cmpMeta, text);
    }
    catch (e) {
        console.log("parseComponentMetaData: " + e);
        console.log(text);
    }
    return null;
}
function updateComponentMeta(cmpMeta, orgText) {
    if (cmpMeta.selector) {
        console.log("Please use \"tag\" instead of \"selector\" in component decorator: " + cmpMeta.selector);
        cmpMeta.tag = cmpMeta.selector;
    }
    if (!cmpMeta.tag || cmpMeta.tag.trim() == '') {
        throw "tag missing in component decorator: " + orgText;
    }
    updateTag(cmpMeta);
    updateModes(cmpMeta);
    updateStyles(cmpMeta);
    updateProperties(cmpMeta);
    return cmpMeta;
}
function updateTag(cmpMeta) {
    cmpMeta.tag = cmpMeta.tag.trim().toLowerCase();
    var invalidChars = cmpMeta.tag.replace(/\w|-/g, '');
    if (invalidChars !== '') {
        throw "\"" + cmpMeta.tag + "\" tag contains invalid characters: " + invalidChars;
    }
    if (cmpMeta.tag.indexOf('-') === -1) {
        throw "\"" + cmpMeta.tag + "\" tag must contain a dash (-)";
    }
    if (cmpMeta.tag.indexOf('-') === 0) {
        throw "\"" + cmpMeta.tag + "\" tag cannot start with a dash (-)";
    }
    if (cmpMeta.tag.lastIndexOf('-') === cmpMeta.tag.length - 1) {
        throw "\"" + cmpMeta.tag + "\" tag cannot end with a dash (-)";
    }
}
function updateModes(cmpMeta) {
    cmpMeta.modes = cmpMeta.modes = {};
}
function updateStyles(cmpMeta) {
    var styleModes = cmpMeta.styleUrls;
    if (styleModes) {
        Object.keys(styleModes).forEach(function (styleModeName) {
            cmpMeta.modes[styleModeName] = {
                styleUrls: [styleModes[styleModeName]]
            };
        });
    }
}
function updateProperties(cmpMeta) {
    if (!cmpMeta.props)
        return;
    var validPropTypes = ['string', 'boolean', 'number', 'Array', 'Object'];
    Object.keys(cmpMeta.props).forEach(function (propName) {
        if (propName.indexOf('-') > -1) {
            throw "\"" + propName + "\" property name cannot have a dash (-) in it";
        }
        if (!isNaN(propName.charAt(0))) {
            throw "\"" + propName + "\" property name cannot start with a number";
        }
        var prop = cmpMeta.props[propName];
        if (prop.type) {
            if (typeof prop.type === 'string') {
                prop.type = prop.type.trim().toLowerCase();
            }
            if (prop.type === 'array') {
                prop.type = 'Array';
            }
            if (prop.type === 'object') {
                prop.type = 'Object';
            }
            if (validPropTypes.indexOf(prop.type) === -1) {
                throw "\"" + propName + "\" invalid for property type: " + prop.type;
            }
        }
    });
}

var IONIC_GLOBALS = ['Component', 'h', 'Ionic', 'Prop', 'Watch'];
function removeImports(ctx) {
    return function (transformContext) {
        function visitImport(importNode) {
            var importSpecifiers = [];
            importNode.importClause.namedBindings.forEachChild(function (nb) {
                if (nb.kind === ts.SyntaxKind.ImportSpecifier) {
                    var importSpecifier = nb;
                    if (IONIC_GLOBALS.indexOf(importSpecifier.name.text) === -1) {
                        importSpecifiers.push(importSpecifier);
                    }
                }
            });
            var namedImports = ts.createNamedImports(importSpecifiers);
            var newImportClause = ts.updateImportClause(importNode.importClause, importNode.importClause.name, namedImports);
            return ts.updateImportDeclaration(importNode, importNode.decorators, importNode.modifiers, newImportClause, importNode.moduleSpecifier);
        }
        function visit(node) {
            switch (node.kind) {
                case ts.SyntaxKind.ImportDeclaration:
                    return visitImport(node);
                default:
                    return ts.visitEachChild(node, function (node) {
                        return visit(node);
                    }, transformContext);
            }
        }
        return function (tsSourceFile) {
            var fileMeta = ctx.files.get(tsSourceFile.fileName);
            if (fileMeta && fileMeta.hasCmpClass) {
                return visit(tsSourceFile);
            }
            return tsSourceFile;
        };
    };
}

function transpile(config, ctx) {
    var tsFileNames = getTsFileNames(ctx);
    if (config.debug) {
        console.log("compile, transpile: " + tsFileNames);
    }
    if (!tsFileNames.length) {
        return Promise.resolve();
    }
    var sourcesMap = new Map();
    ctx.files.forEach(function (fileMeta, filePath) {
        var sourceFile = ts.createSourceFile(filePath, fileMeta.srcText, ts.ScriptTarget.ES2015);
        sourcesMap.set(filePath, sourceFile);
    });
    var outputs = new Map();
    var tsCompilerOptions = createTsCompilerConfigs(config);
    var tsHost = {
        getSourceFile: function (filePath) { return sourcesMap.get(filePath); },
        getDefaultLibFileName: function () { return 'lib.d.ts'; },
        getCurrentDirectory: function () { return ''; },
        getDirectories: function () { return []; },
        getCanonicalFileName: function (fileName) { return fileName; },
        useCaseSensitiveFileNames: function () { return true; },
        getNewLine: function () { return '\n'; },
        fileExists: function (filePath) {
            return ctx.files.has(filePath);
        },
        readFile: function (filePath) {
            var fileMeta = ctx.files.get(filePath);
            if (fileMeta) {
                return fileMeta.srcText;
            }
            fileMeta = createFileMeta(config.packages, ctx, filePath, config.packages.fs.readFileSync(filePath, 'utf-8'));
            return fileMeta.srcText;
        },
        writeFile: function (jsFilePath, jsText, writeByteOrderMark, onError, sourceFiles) {
            sourceFiles.forEach(function (s) {
                var fileMeta = ctx.files.get(s.fileName);
                if (fileMeta) {
                    fileMeta.jsFilePath = jsFilePath;
                    fileMeta.jsText = jsText;
                }
            });
            if (jsText && jsText.trim().length) {
                outputs.set(jsFilePath, jsText);
            }
            writeByteOrderMark;
            onError;
        }
    };
    var program = ts.createProgram(tsFileNames, tsCompilerOptions, tsHost);
    if (program.getSyntacticDiagnostics().length > 0) {
        return Promise.reject(program.getSyntacticDiagnostics());
    }
    var result = program.emit(undefined, tsHost.writeFile, undefined, false, {
        before: [
            componentClass(ctx),
            removeImports(ctx)
        ]
    });
    if (result.diagnostics.length > 0) {
        return Promise.reject(result.diagnostics);
    }
    return writeFiles(config.packages, outputs);
}
function createTsCompilerConfigs(config) {
    var tsCompilerOptions = Object.assign({}, config.compilerOptions);
    tsCompilerOptions.noImplicitUseStrict = true;
    tsCompilerOptions.moduleResolution = ts.ModuleResolutionKind.NodeJs;
    tsCompilerOptions.module = ts.ModuleKind.ES2015;
    tsCompilerOptions.target = getTsScriptTarget(config.compilerOptions.target);
    tsCompilerOptions.isolatedModules = true;
    tsCompilerOptions.lib = tsCompilerOptions.lib || [];
    if (!tsCompilerOptions.lib.indexOf('lib.dom.d.ts')) {
        tsCompilerOptions.lib.push('lib.dom.d.ts');
    }
    if (!tsCompilerOptions.lib.indexOf('lib.es2015.d.ts')) {
        tsCompilerOptions.lib.push('lib.es2015.d.ts');
    }
    if (!tsCompilerOptions.lib.indexOf('lib.es5.d.ts')) {
        tsCompilerOptions.lib.push('lib.es5.d.ts');
    }
    return tsCompilerOptions;
}
function getTsFileNames(ctx) {
    var fileNames = [];
    ctx.files.forEach(function (fileMeta) {
        if (fileMeta.isTsSourceFile && fileMeta.hasCmpClass) {
            fileNames.push(fileMeta.filePath);
        }
    });
    return fileNames;
}
function getTsScriptTarget(str) {
    if (str === 'es2015') {
        return ts.ScriptTarget.ES2015;
    }
    return ts.ScriptTarget.ES5;
}

function compile(config, ctx) {
    if (ctx === void 0) { ctx = {}; }
    if (!config.packages) {
        throw 'config.packages required';
    }
    if (!config.packages.fs) {
        throw 'config.packages.fs required';
    }
    if (!config.packages.path) {
        throw 'config.packages.path required';
    }
    if (!config.packages.nodeSass) {
        throw 'config.packages.nodeSass required';
    }
    if (!config.packages.rollup) {
        throw 'config.packages.rollup required';
    }
    if (!config.packages.typescript) {
        throw 'config.packages.typescript required';
    }
    if (config.debug) {
        console.log("compile, include: " + config.include);
        console.log("compile, outDir: " + config.compilerOptions.outDir);
    }
    if (!ctx.files) {
        ctx.files = new Map();
    }
    ctx.results = {};
    config.include = config.include || [];
    if (!config.exclude) {
        config.exclude = ['node_modules', 'bower_components'];
    }
    var promises = config.include.map(function (includePath) {
        return scanDirectory(includePath, config, ctx);
    });
    return Promise.all(promises)
        .then(function () {
        return transpile(config, ctx);
    }).then(function () {
        return processStyles(config, ctx);
    }).then(function () {
        return generateManifest(config, ctx);
    }).then(function () {
        return ctx.results;
    });
}
function scanDirectory(dir, config, ctx) {
    return new Promise(function (resolve) {
        if (config.debug) {
            console.log("compile, scanDirectory: " + dir);
        }
        config.packages.fs.readdir(dir, function (err, files) {
            if (err) {
                logError(ctx.results, err);
                resolve();
                return;
            }
            var promises = [];
            files.forEach(function (dirItem) {
                var readPath = config.packages.path.join(dir, dirItem);
                if (!isValidDirectory(config, readPath)) {
                    return;
                }
                promises.push(new Promise(function (resolve) {
                    config.packages.fs.stat(readPath, function (err, stats) {
                        if (err) {
                            logError(ctx.results, err);
                            resolve();
                        }
                        else if (stats.isDirectory()) {
                            scanDirectory(readPath, config, ctx).then(function () {
                                resolve();
                            });
                        }
                        else if (isTsSourceFile(readPath)) {
                            getFileMeta(config, ctx, readPath).then(function () {
                                resolve();
                            });
                        }
                        else {
                            resolve();
                        }
                    });
                }));
            });
            Promise.all(promises).then(function () {
                resolve();
            });
        });
    });
}
function isValidDirectory(config, filePath) {
    for (var i = 0; i < config.exclude.length; i++) {
        if (filePath.indexOf(config.exclude[i]) > -1) {
            return false;
        }
    }
    return true;
}
function processStyles(config, ctx) {
    var destDir = config.compilerOptions.outDir;
    var promises = [];
    var includedSassFiles = [];
    ctx.files.forEach(function (f) {
        if (!f.isTsSourceFile || !f.cmpMeta)
            return;
        Object.keys(f.cmpMeta.modes).forEach(function (modeName) {
            f.cmpMeta.modes[modeName].styleUrls.forEach(function (styleUrl) {
                var srcAbsolutePath = config.packages.path.join(f.srcDir, styleUrl);
                promises.push(getIncludedSassFiles(config, ctx, includedSassFiles, srcAbsolutePath));
            });
        });
    });
    return Promise.all(promises).then(function () {
        var files = new Map();
        var promises = [];
        includedSassFiles.forEach(function (includedSassFile) {
            config.include.forEach(function (includeDir) {
                if (includedSassFile.indexOf(includeDir) === 0) {
                    var src = includedSassFile;
                    var relative = includedSassFile.replace(includeDir, '');
                    var dest_1 = config.packages.path.join(destDir, relative);
                    promises.push(readFile(config.packages, src).then(function (content) {
                        files.set(dest_1, content);
                    }));
                }
            });
        });
        return Promise.all(promises).then(function () {
            return writeFiles(config.packages, files);
        });
    });
}
function getIncludedSassFiles(config, ctx, includedSassFiles, scssFilePath) {
    return new Promise(function (resolve, reject) {
        var sassConfig = {
            file: scssFilePath
        };
        if (config.debug) {
            console.log("compile, getIncludedSassFiles: " + scssFilePath);
        }
        config.packages.nodeSass.render(sassConfig, function (err, result) {
            if (err) {
                logError(ctx.results, err);
                reject(err);
            }
            else {
                result.stats.includedFiles.forEach(function (includedFile) {
                    if (includedSassFiles.indexOf(includedFile) === -1) {
                        includedSassFiles.push(includedFile);
                    }
                });
                resolve();
            }
        });
    });
}
function generateManifest(config, ctx) {
    var manifest = {
        components: {},
        bundles: []
    };
    var components = [];
    var destDir = config.compilerOptions.outDir;
    ctx.files.forEach(function (f) {
        if (!f.isTsSourceFile || !f.cmpMeta)
            return;
        var componentUrl = f.jsFilePath.replace(destDir + config.packages.path.sep, '');
        var modes = f.cmpMeta.modes;
        var componentDir = config.packages.path.dirname(componentUrl);
        Object.keys(modes).forEach(function (modeName) {
            modes[modeName].styleUrls = modes[modeName].styleUrls.map(function (styleUrl) {
                return config.packages.path.join(componentDir, styleUrl);
            });
        });
        components.push({
            tag: f.cmpMeta.tag,
            componentUrl: componentUrl,
            componentClass: f.cmpClassName,
            modes: modes,
            props: f.cmpMeta.props || {},
            watches: f.cmpMeta.watches || {}
        });
    });
    if (config.bundles) {
        manifest.bundles = config.bundles;
    }
    else {
        ctx.files.forEach(function (f) {
            if (f.isTsSourceFile && f.cmpMeta) {
                manifest.bundles.push([f.cmpMeta.tag]);
            }
        });
    }
    components = components.sort(function (a, b) {
        if (a.tag < b.tag) {
            return -1;
        }
        if (a.tag > b.tag) {
            return 1;
        }
        return 0;
    });
    components.forEach(function (cmp) {
        manifest.components[cmp.tag] = {
            componentUrl: cmp.componentUrl,
            componentClass: cmp.componentClass,
            modes: cmp.modes,
            props: cmp.props,
            watches: cmp.watches
        };
    });
    var manifestFile = config.packages.path.join(config.compilerOptions.outDir, 'manifest.json');
    var json = JSON.stringify(manifest, null, 2);
    return writeFile(config.packages, manifestFile, json);
}

exports.bundle = bundle;
exports.compile = compile;
