(function (window, document) {
    "use strict";

    function vnode(sel, data, children, text, elm) {
        var key = data === undefined ? undefined : data.key;
        return { sel: sel, vdata: data, vchildren: children,
            vtext: text, elm: elm, vkey: key };
    }

    function isDef(s) {
        return s !== undefined && s !== null;
    }
    function isUndef(s) {
        return s === undefined;
    }
    function isArray(val) {
        return !!val && val.constructor === Array;
    }
    function isObject(val) {
        return typeof val === 'object';
    }

    function isString(val) {
        return typeof val === 'string';
    }
    function isNumber(val) {
        return typeof val === 'number';
    }
    function isFunction(val) {
        return typeof val === 'function';
    }
    function isStringOrNumber(s) {
        return isString(s) || isNumber(s);
    }
    function toCamelCase(str) {
        return str.replace(/-([a-z])/g, function (g) {
            return g[1].toUpperCase();
        });
    }
    function toDashCase(str) {
        return str.replace(/([A-Z])/g, function (g) {
            return '-' + g[0].toLowerCase();
        });
    }

    function addNS(data, children, sel) {
        data.ns = 'http://www.w3.org/2000/svg';
        if (sel !== 'foreignObject' && children !== undefined) {
            for (var i = 0; i < children.length; ++i) {
                var childData = children[i].vdata;
                if (childData !== undefined) {
                    addNS(childData, children[i].vchildren, children[i].sel);
                }
            }
        }
    }
    function h(sel, b, c) {
        var data = {},
            children,
            text,
            i;
        var elm = undefined;
        if (sel.nodeType) {
            elm = sel;
        }
        if (c !== undefined) {
            data = b;
            if (isArray(c)) {
                children = c;
            } else if (isStringOrNumber(c)) {
                text = c;
            } else if (c && c.sel) {
                children = [c];
            }
        } else if (b !== undefined) {
            if (isArray(b)) {
                children = b;
            } else if (isStringOrNumber(b)) {
                text = b;
            } else if (b && b.sel) {
                children = [b];
            } else {
                data = b;
            }
        }
        if (isArray(children)) {
            for (i = 0; i < children.length; ++i) {
                if (isStringOrNumber(children[i])) children[i] = vnode(undefined, undefined, undefined, children[i]);
            }
        }
        if (sel[0] === 's' && sel[1] === 'v' && sel[2] === 'g' && (sel.length === 3 || sel[3] === '.' || sel[3] === '#')) {
            addNS(data, children, sel);
        }
        return vnode(sel, data, children, text, elm);
    }

    var xlinkNS = 'http://www.w3.org/1999/xlink';
    var xmlNS = 'http://www.w3.org/XML/1998/namespace';
    function updateAttrs(oldVnode, vnode) {
        var key,
            cur,
            old,
            elm = vnode.elm,
            oldAttrs = oldVnode.vdata.attrs,
            attrs = vnode.vdata.attrs;
        if (!oldAttrs && !attrs) return;
        if (oldAttrs === attrs) return;
        oldAttrs = oldAttrs || {};
        attrs = attrs || {};
        // update modified attributes, add new attributes
        for (key in attrs) {
            cur = attrs[key];
            old = oldAttrs[key];
            if (old !== cur) {
                if (typeof cur === 'boolean') {
                    if (cur) {
                        elm.setAttribute(key, "");
                    } else {
                        elm.removeAttribute(key);
                    }
                } else {
                    if (key.charCodeAt(0) !== 120) {
                        elm.setAttribute(key, cur);
                    } else if (key.charCodeAt(3) === 58) {
                        // Assume xml namespace
                        elm.setAttributeNS(xmlNS, key, cur);
                    } else if (key.charCodeAt(5) === 58) {
                        // Assume xlink namespace
                        elm.setAttributeNS(xlinkNS, key, cur);
                    } else {
                        elm.setAttribute(key, cur);
                    }
                }
            }
        }
        // remove removed attributes
        // use `in` operator since the previous `for` iteration uses it (.i.e. add even attributes with undefined value)
        // the other option is to remove all attributes with value == undefined
        for (key in oldAttrs) {
            if (!(key in attrs)) {
                elm.removeAttribute(key);
            }
        }
    }

    function updateClass(oldVnode, vnode) {
        var cur,
            name,
            elm = vnode.elm,
            oldClass = oldVnode.vdata.class,
            klass = vnode.vdata.class;
        if (!oldClass && !klass) return;
        if (oldClass === klass) return;
        oldClass = oldClass || {};
        klass = klass || {};
        for (name in oldClass) {
            if (!klass[name]) {
                elm.classList.remove(name);
            }
        }
        for (name in klass) {
            cur = klass[name];
            if (cur !== oldClass[name]) {
                elm.classList[cur ? 'add' : 'remove'](name);
            }
        }
    }

    function invokeHandler(handler, vnode, event) {
        if (isFunction(handler)) {
            // call function handler
            handler.call(vnode, event, vnode);
        } else if (isObject(handler)) {
            // call handler with arguments
            if (isFunction(handler[0])) {
                // special case for single argument for performance
                if (handler.length === 2) {
                    handler[0].call(vnode, handler[1], event, vnode);
                } else {
                    var args = handler.slice(1);
                    args.push(event);
                    args.push(vnode);
                    handler[0].apply(vnode, args);
                }
            } else {
                // call multiple handlers
                for (var i = 0; i < handler.length; i++) {
                    invokeHandler(handler[i]);
                }
            }
        }
    }
    function handleEvent(event, vnode) {
        var name = event.type,
            on = vnode.vdata.on;
        // call event handler(s) if exists
        if (on && on[name]) {
            invokeHandler(on[name], vnode, event);
        }
    }
    function createListener() {
        return function handler(event) {
            handleEvent(event, handler.vnode);
        };
    }
    function updateEventListeners(oldVnode, vnode) {
        var oldOn = oldVnode.vdata.on,
            oldListener = oldVnode.listener,
            oldElm = oldVnode.elm,
            on = vnode && vnode.vdata.on,
            elm = vnode && vnode.elm,
            name;
        // optimization for reused immutable handlers
        if (oldOn === on) {
            return;
        }
        // remove existing listeners which no longer used
        if (oldOn && oldListener) {
            // if element changed or deleted we remove all existing listeners unconditionally
            if (!on) {
                for (name in oldOn) {
                    // remove listener if element was changed or existing listeners removed
                    oldElm.removeEventListener(name, oldListener, false);
                }
            } else {
                for (name in oldOn) {
                    // remove listener if existing listener removed
                    if (!on[name]) {
                        oldElm.removeEventListener(name, oldListener, false);
                    }
                }
            }
        }
        // add new listeners which has not already attached
        if (on) {
            // reuse existing listener or create new
            var listener = vnode.listener = oldVnode.listener || createListener();
            // update vnode for listener
            listener.vnode = vnode;
            // if element changed or added we add all needed listeners unconditionally
            if (!oldOn) {
                for (name in on) {
                    // add listener if element was changed or new listeners added
                    elm.addEventListener(name, listener, false);
                }
            } else {
                for (name in on) {
                    // add listener if new listener added
                    if (!oldOn[name]) {
                        elm.addEventListener(name, listener, false);
                    }
                }
            }
        }
    }

    /**
     * Virtual DOM patching algorithm based on Snabbdom by
     * Simon Friis Vindum (@paldepind)
     * Licensed under the MIT License
     * https://github.com/snabbdom/snabbdom/blob/master/LICENSE
     *
     * Modified for Ionic's Web Component Renderer
     * /
    
    
    /* global module, document, Node */
    var emptyNode = vnode('', {}, [], undefined, undefined);
    function sameVnode(vnode1, vnode2) {
        return vnode1.vkey === vnode2.vkey && vnode1.sel === vnode2.sel;
    }
    function isVnode(vnode$$1) {
        return vnode$$1.sel !== undefined || vnode$$1.elm !== undefined;
    }
    function createKeyToOldIdx(children, beginIdx, endIdx) {
        var i = void 0,
            map = {},
            key = void 0,
            ch = void 0;
        for (i = beginIdx; i <= endIdx; ++i) {
            ch = children[i];
            if (ch != null) {
                key = ch.vkey;
                if (key !== undefined) map[key] = i;
            }
        }
        return map;
    }
    function Renderer(api) {
        function emptyNodeAt(elm) {
            var id = elm.id ? '#' + elm.id : '';
            var c = elm.className ? '.' + elm.className.split(' ').join('.') : '';
            return vnode(api.$tagName(elm) + id + c, {}, [], undefined, elm);
        }
        function createRmCb(childElm) {
            return function rmCb() {
                var parent = api.$parentNode(childElm);
                api.$removeChild(parent, childElm);
            };
        }
        function createElm(vnode$$1, insertedVnodeQueue) {
            var i = void 0,
                data = vnode$$1.vdata;
            if (data !== undefined) {
                if (isDef(i = data.hook) && isDef(i = i.init)) {
                    i(vnode$$1);
                    data = vnode$$1.vdata;
                }
            }
            var children = vnode$$1.vchildren,
                sel = vnode$$1.sel;
            if (sel === '!') {
                if (isUndef(vnode$$1.vtext)) {
                    vnode$$1.vtext = '';
                }
                vnode$$1.elm = api.$createComment(vnode$$1.vtext);
            } else if (sel !== undefined) {
                // Parse selector
                var hashIdx = sel.indexOf('#');
                var dotIdx = sel.indexOf('.', hashIdx);
                var hash = hashIdx > 0 ? hashIdx : sel.length;
                var dot = dotIdx > 0 ? dotIdx : sel.length;
                var tag = hashIdx !== -1 || dotIdx !== -1 ? sel.slice(0, Math.min(hash, dot)) : sel;
                var elm = vnode$$1.elm = isDef(data) && isDef(i = data.ns) ? api.$createElementNS(i, tag) : api.$createElement(tag);
                if (hash < dot) elm.id = sel.slice(hash + 1, dot);
                if (dotIdx > 0) elm.className = sel.slice(dot + 1).replace(/\./g, ' ');
                updateAttrs(emptyNode, vnode$$1);
                updateClass(emptyNode, vnode$$1);
                updateEventListeners(emptyNode, vnode$$1);
                if (isArray(children)) {
                    for (i = 0; i < children.length; ++i) {
                        var ch = children[i];
                        if (ch != null) {
                            api.$appendChild(elm, createElm(ch, insertedVnodeQueue));
                        }
                    }
                } else if (isStringOrNumber(vnode$$1.vtext)) {
                    api.$appendChild(elm, api.$createTextNode(vnode$$1.vtext));
                }
                i = vnode$$1.vdata.hook; // Reuse variable
                if (isDef(i)) {
                    if (i.create) i.create(emptyNode, vnode$$1);
                    if (i.insert) insertedVnodeQueue.push(vnode$$1);
                }
            } else {
                vnode$$1.elm = api.$createTextNode(vnode$$1.vtext);
            }
            return vnode$$1.elm;
        }
        function addVnodes(parentElm, before, vnodes, startIdx, endIdx, insertedVnodeQueue) {
            for (; startIdx <= endIdx; ++startIdx) {
                var ch = vnodes[startIdx];
                if (ch != null) {
                    api.$insertBefore(parentElm, createElm(ch, insertedVnodeQueue), before);
                }
            }
        }
        function removeVnodes(parentElm, vnodes, startIdx, endIdx) {
            for (; startIdx <= endIdx; ++startIdx) {
                var i = void 0,
                    rm = void 0,
                    ch = vnodes[startIdx];
                if (ch != null) {
                    if (isDef(ch.sel)) {
                        invokeDestroyHook(ch);
                        rm = createRmCb(ch.elm);
                        if (isDef(i = ch.vdata) && isDef(i = i.hook) && isDef(i = i.remove)) {
                            i(ch, rm);
                        } else {
                            rm();
                        }
                    } else {
                        api.$removeChild(parentElm, ch.elm);
                    }
                }
            }
        }
        function updateChildren(parentElm, oldCh, newCh, insertedVnodeQueue) {
            var oldStartIdx = 0,
                newStartIdx = 0;
            var oldEndIdx = oldCh.length - 1;
            var oldStartVnode = oldCh[0];
            var oldEndVnode = oldCh[oldEndIdx];
            var newEndIdx = newCh.length - 1;
            var newStartVnode = newCh[0];
            var newEndVnode = newCh[newEndIdx];
            var oldKeyToIdx = void 0;
            var idxInOld = void 0;
            var elmToMove = void 0;
            var before = void 0;
            while (oldStartIdx <= oldEndIdx && newStartIdx <= newEndIdx) {
                if (oldStartVnode == null) {
                    oldStartVnode = oldCh[++oldStartIdx]; // Vnode might have been moved left
                } else if (oldEndVnode == null) {
                    oldEndVnode = oldCh[--oldEndIdx];
                } else if (newStartVnode == null) {
                    newStartVnode = newCh[++newStartIdx];
                } else if (newEndVnode == null) {
                    newEndVnode = newCh[--newEndIdx];
                } else if (sameVnode(oldStartVnode, newStartVnode)) {
                    patchVnode(oldStartVnode, newStartVnode, insertedVnodeQueue);
                    oldStartVnode = oldCh[++oldStartIdx];
                    newStartVnode = newCh[++newStartIdx];
                } else if (sameVnode(oldEndVnode, newEndVnode)) {
                    patchVnode(oldEndVnode, newEndVnode, insertedVnodeQueue);
                    oldEndVnode = oldCh[--oldEndIdx];
                    newEndVnode = newCh[--newEndIdx];
                } else if (sameVnode(oldStartVnode, newEndVnode)) {
                    patchVnode(oldStartVnode, newEndVnode, insertedVnodeQueue);
                    api.$insertBefore(parentElm, oldStartVnode.elm, api.$nextSibling(oldEndVnode.elm));
                    oldStartVnode = oldCh[++oldStartIdx];
                    newEndVnode = newCh[--newEndIdx];
                } else if (sameVnode(oldEndVnode, newStartVnode)) {
                    patchVnode(oldEndVnode, newStartVnode, insertedVnodeQueue);
                    api.$insertBefore(parentElm, oldEndVnode.elm, oldStartVnode.elm);
                    oldEndVnode = oldCh[--oldEndIdx];
                    newStartVnode = newCh[++newStartIdx];
                } else {
                    if (oldKeyToIdx === undefined) {
                        oldKeyToIdx = createKeyToOldIdx(oldCh, oldStartIdx, oldEndIdx);
                    }
                    idxInOld = oldKeyToIdx[newStartVnode.vkey];
                    if (isUndef(idxInOld)) {
                        api.$insertBefore(parentElm, createElm(newStartVnode, insertedVnodeQueue), oldStartVnode.elm);
                        newStartVnode = newCh[++newStartIdx];
                    } else {
                        elmToMove = oldCh[idxInOld];
                        if (elmToMove.sel !== newStartVnode.sel) {
                            api.$insertBefore(parentElm, createElm(newStartVnode, insertedVnodeQueue), oldStartVnode.elm);
                        } else {
                            patchVnode(elmToMove, newStartVnode, insertedVnodeQueue);
                            oldCh[idxInOld] = undefined;
                            api.$insertBefore(parentElm, elmToMove.elm, oldStartVnode.elm);
                        }
                        newStartVnode = newCh[++newStartIdx];
                    }
                }
            }
            if (oldStartIdx > oldEndIdx) {
                before = newCh[newEndIdx + 1] == null ? null : newCh[newEndIdx + 1].elm;
                addVnodes(parentElm, before, newCh, newStartIdx, newEndIdx, insertedVnodeQueue);
            } else if (newStartIdx > newEndIdx) {
                removeVnodes(parentElm, oldCh, oldStartIdx, oldEndIdx);
            }
        }
        function patchVnode(oldVnode, vnode$$1, insertedVnodeQueue) {
            var i = void 0,
                hook = void 0;
            if (isDef(i = vnode$$1.vdata) && isDef(hook = i.hook) && isDef(i = hook.prepatch)) {
                i(oldVnode, vnode$$1);
            }
            var elm = vnode$$1.elm = oldVnode.elm;
            var oldCh = oldVnode.vchildren;
            var ch = vnode$$1.vchildren;
            if (oldVnode === vnode$$1) return;
            if (vnode$$1.vdata !== undefined) {
                updateAttrs(oldVnode, vnode$$1);
                updateClass(oldVnode, vnode$$1);
                updateEventListeners(emptyNode, vnode$$1);
                i = vnode$$1.vdata.hook;
                if (isDef(i) && isDef(i = i.update)) i(oldVnode, vnode$$1);
            }
            if (isUndef(vnode$$1.vtext)) {
                if (isDef(oldCh) && isDef(ch)) {
                    if (oldCh !== ch) {
                        updateChildren(elm.shadowRoot || elm, oldCh, ch, insertedVnodeQueue);
                    }
                } else if (isDef(ch)) {
                    if (isDef(oldVnode.vtext)) api.$setTextContent(elm, '');
                    addVnodes(elm, null, ch, 0, ch.length - 1, insertedVnodeQueue);
                } else if (isDef(oldCh)) {
                    removeVnodes(elm, oldCh, 0, oldCh.length - 1);
                } else if (isDef(oldVnode.vtext)) {
                    api.$setTextContent(elm, '');
                }
            } else if (oldVnode.vtext !== vnode$$1.vtext) {
                api.$setTextContent(elm, vnode$$1.vtext);
            }
            if (isDef(hook) && isDef(i = hook.postpatch)) {
                i(oldVnode, vnode$$1);
            }
        }
        return function patch(oldVnode, vnode$$1) {
            var elm = void 0,
                parent = void 0;
            var insertedVnodeQueue = [];
            if (!isVnode(oldVnode)) {
                oldVnode = emptyNodeAt(oldVnode);
            }
            if (vnode$$1.elm || sameVnode(oldVnode, vnode$$1)) {
                patchVnode(oldVnode, vnode$$1, insertedVnodeQueue);
            } else {
                elm = oldVnode.elm;
                parent = api.$parentNode(elm);
                createElm(vnode$$1, insertedVnodeQueue);
                if (parent !== null) {
                    api.$insertBefore(parent, vnode$$1.elm, api.$nextSibling(elm));
                    removeVnodes(parent, [oldVnode], 0, 0);
                }
            }
            return vnode$$1;
        };
    }
    function invokeDestroyHook(vnode$$1) {
        var i = void 0,
            j = void 0,
            data = vnode$$1.vdata;
        if (data !== undefined) {
            updateEventListeners(vnode$$1);
            if (vnode$$1.vchildren !== undefined) {
                for (j = 0; j < vnode$$1.vchildren.length; ++j) {
                    i = vnode$$1.vchildren[j];
                    if (i != null && !isString(i)) {
                        invokeDestroyHook(i);
                    }
                }
            }
        }
    }

    function generateVNode(elm, instance, hostCss) {
        var vnode$$1 = instance.render && instance.render();
        if (vnode$$1) {
            vnode$$1.elm = elm;
        } else {
            // use the default render function instead
            vnode$$1 = h(elm, h('div', themeVNodeData(instance, hostCss), h('slot')));
        }
        delete vnode$$1.sel;
        return vnode$$1;
    }
    function themeVNodeData(instance, cssClassName) {
        var vnodeData = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

        var cssClasses = vnodeData.class = vnodeData.class || {};
        var mode = instance.mode;
        var color = instance.color;
        cssClasses[cssClassName] = true;
        if (mode) {
            cssClasses[cssClassName + '-' + mode] = true;
            if (color) {
                cssClasses[cssClassName + '-' + color] = cssClasses[cssClassName + '-' + mode + '-' + color] = true;
            }
        }
        return vnodeData;
    }

    function PlatformClient(win, doc, ionic, staticDir, domCtrl, nextTickCtrl) {
        var registry = {};
        var loadedBundles = {};
        var bundleCallbacks = {};
        var activeJsonRequests = {};
        var moduleImports = {};
        var css = {};
        var hasNativeShadowDom = !(win.ShadyDOM && win.ShadyDOM.inUse);
        var injectedIonic = {
            theme: themeVNodeData,
            emit: function emitEvent(instance, eventName, data) {
                var ev = doc.createEvent('CustomEvent');
                ev.initCustomEvent(eventName, true, true, data);
                instance.$el.dispatchEvent(ev);
            }
        };
        ionic.loadComponents = function loadComponents(bundleId) {
            var args = arguments;
            for (var i = 1; i < args.length; i++) {
                // first arg is the bundleId
                // each arg after that is a component/mode
                var cmpModeData = args[i];
                // tag name (ion-badge)
                // get component meta data by tag name
                var cmpMeta = registry[cmpModeData[0]];
                // component instance property watches
                cmpMeta.watches = cmpModeData[2];
                // mode name (ios, md, wp)
                // get component mode
                var cmpMode = cmpMeta.modes[cmpModeData[3]];
                if (cmpMode) {
                    // component mode styles
                    cmpMode.styles = cmpModeData[4];
                }
                // import component function
                // inject ionic globals
                cmpModeData[5](moduleImports, h, injectedIonic);
                // get the component class which was added to moduleImports
                // component class name (Badge)
                cmpMeta.componentModule = moduleImports[cmpModeData[1]];
                // fire off all the callbacks waiting on this bundle to load
                var callbacks = bundleCallbacks[bundleId];
                if (callbacks) {
                    for (var j = 0, jlen = callbacks.length; j < jlen; j++) {
                        callbacks[j]();
                    }
                    delete bundleCallbacks[bundleId];
                }
                // remember that we've already loaded this bundle
                loadedBundles[bundleId] = true;
            }
        };
        function loadComponent(bundleId, cb) {
            if (loadedBundles[bundleId]) {
                // we've already loaded this bundle
                cb();
            } else {
                // never seen this bundle before, let's start the request
                // and add it to the bundle callbacks to fire when it's loaded
                if (bundleCallbacks[bundleId]) {
                    bundleCallbacks[bundleId].push(cb);
                } else {
                    bundleCallbacks[bundleId] = [cb];
                }
                // create the url we'll be requesting
                var url = staticDir + 'ionic.' + bundleId + '.js';
                if (!activeJsonRequests[url]) {
                    // not already actively requesting this url
                    // let's kick off the request
                    jsonp(url);
                }
            }
        }
        function jsonp(url) {
            // remember that we're actively requesting this url
            activeJsonRequests[url] = true;
            // create a sript element to add to the document.head
            var scriptElm = createElement('script');
            scriptElm.charset = 'utf-8';
            scriptElm.async = true;
            scriptElm.src = url;
            // create a fallback timeout if something goes wrong
            var tmrId = setTimeout(onScriptComplete, 120000);
            function onScriptComplete() {
                clearTimeout(tmrId);
                scriptElm.onerror = scriptElm.onload = null;
                scriptElm.parentNode.removeChild(scriptElm);
                // remove from our list of active requests
                delete activeJsonRequests[url];
            }
            // add script completed listener to this script element
            scriptElm.onerror = scriptElm.onload = onScriptComplete;
            // inject a script tag in the head
            // kick off the actual request
            doc.head.appendChild(scriptElm);
        }
        function attachShadow(elm, cmpMode, cmpModeId) {
            var shadowElm = elm.attachShadow({ mode: 'open' });
            if (hasNativeShadowDom) {
                if (!cmpMode.styleElm) {
                    cmpMode.styleElm = createElement('style');
                    cmpMode.styleElm.innerHTML = cmpMode.styles;
                }
                shadowElm.appendChild(cmpMode.styleElm.cloneNode(true));
            } else if (!hasCss(cmpModeId)) {
                var headStyleEle = createElement('style');
                headStyleEle.dataset['cmpModeId'] = cmpModeId;
                headStyleEle.innerHTML = cmpMode.styles.replace(/\:host\-context\((.*?)\)|:host\((.*?)\)|\:host/g, '__h');
                appendChild(doc.head, headStyleEle);
                setCss(cmpModeId);
            }
            return shadowElm;
        }
        function registerComponent(tag, data) {
            var modeBundleIds = data[0];
            var props = data[1] || {};
            var cmpMeta = registry[tag] = {
                tag: tag,
                modes: {},
                props: props,
                obsAttrs: []
            };
            var keys = Object.keys(modeBundleIds);
            for (var i = 0; i < keys.length; i++) {
                cmpMeta.modes[keys[i]] = {
                    bundleId: modeBundleIds[keys[i]]
                };
            }
            keys = cmpMeta.tag.split('-');
            keys.shift();
            cmpMeta.hostCss = keys.join('-');
            props.color = {};
            props.mode = {};
            keys = Object.keys(props);
            for (i = 0; i < keys.length; i++) {
                cmpMeta.obsAttrs.push(toDashCase(keys[i]));
            }
            return cmpMeta;
        }
        function getComponentMeta(tag) {
            return registry[tag];
        }
        function createElement(tagName) {
            return doc.createElement(tagName);
        }
        function createElementNS(namespaceURI, qualifiedName) {
            return doc.createElementNS(namespaceURI, qualifiedName);
        }
        function createTextNode(text) {
            return doc.createTextNode(text);
        }
        function createComment(text) {
            return doc.createComment(text);
        }
        function insertBefore(parentNode, newNode, referenceNode) {
            parentNode.insertBefore(newNode, referenceNode);
        }
        function removeChild(parentNode, childNode) {
            parentNode.removeChild(childNode);
        }
        function appendChild(parentNode, childNode) {
            parentNode.appendChild(childNode);
        }
        function parentNode(node) {
            return node.parentNode;
        }
        function nextSibling(node) {
            return node.nextSibling;
        }
        function tagName(elm) {
            return (elm.tagName || '').toLowerCase();
        }
        function setTextContent(node, text) {
            node.textContent = text;
        }
        function getTextContent(node) {
            return node.textContent;
        }
        function getAttribute(elm, attrName) {
            return elm.getAttribute(attrName);
        }
        function isElement(node) {
            return node.nodeType === 1;
        }
        function isText(node) {
            return node.nodeType === 3;
        }
        function isComment(node) {
            return node.nodeType === 8;
        }
        function hasCss(moduleId) {
            return !!css[moduleId];
        }
        function setCss(linkUrl) {
            css[linkUrl] = true;
        }
        return {
            registerComponent: registerComponent,
            getComponentMeta: getComponentMeta,
            loadComponent: loadComponent,
            isElement: isElement,
            isText: isText,
            isComment: isComment,
            nextTick: nextTickCtrl.nextTick.bind(nextTickCtrl),
            domRead: domCtrl.read.bind(domCtrl),
            domWrite: domCtrl.write.bind(domCtrl),
            $createElement: createElement,
            $createElementNS: createElementNS,
            $createTextNode: createTextNode,
            $createComment: createComment,
            $insertBefore: insertBefore,
            $removeChild: removeChild,
            $appendChild: appendChild,
            $parentNode: parentNode,
            $nextSibling: nextSibling,
            $tagName: tagName,
            $setTextContent: setTextContent,
            $getTextContent: getTextContent,
            $getAttribute: getAttribute,
            $attachShadow: attachShadow
        };
    }

    function attributeChangedCallback(elm, cmpMeta, attrName, oldVal, newVal) {
        if (oldVal !== newVal) {
            var propName = toCamelCase(attrName);
            var prop = cmpMeta.props[propName];
            if (prop) {
                if (prop.type === 'boolean') {
                    elm[propName] = newVal === null || newVal === 'false' ? false : true;
                } else if (prop.type === 'number') {
                    elm[propName] = parseFloat(newVal);
                } else {
                    elm[propName] = newVal;
                }
            }
        }
    }

    function initProps(plt, config, renderer, elm, tag, instance, props, watches) {
        var propValues = {};
        Object.keys(props).forEach(function (propName) {
            var watcher = watches[propName] ? instance[watches[propName].fn].bind(instance) : null;
            propValues[propName] = getInitialValue(config, elm, props[propName].type, propName);
            function getPropValue() {
                return propValues[propName];
            }
            function setPropValue(value) {
                if (propValues[propName] !== value) {
                    propValues[propName] = value;
                    watcher && watcher(value);
                    queueUpdate(plt, config, renderer, elm, tag);
                }
            }
            // dom's element instance
            Object.defineProperty(elm, propName, {
                configurable: true,
                get: getPropValue,
                set: setPropValue
            });
            // user's component class instance
            Object.defineProperty(instance, propName, {
                configurable: true,
                get: getPropValue,
                set: setPropValue
            });
        });
    }
    function getInitialValue(config, elm, propType, propName) {
        if (elm[propName] !== undefined) {
            return elm[propName];
        }
        if (propType === 'boolean') {
            return config.getBoolean(propName);
        }
        if (propType === 'number') {
            return config.getNumber(propName);
        }
        return config.get(propName);
    }

    function queueUpdate(plt, config, renderer, elm, tag) {
        // only run patch if it isn't queued already
        if (!elm.$queued) {
            elm.$queued = true;
            // run the patch in the next tick
            plt.nextTick(function queueUpdateNextTick() {
                // vdom diff and patch the host element for differences
                update(plt, config, renderer, elm, tag);
                // no longer queued
                elm.$queued = false;
            });
        }
    }
    function update(plt, config, renderer, elm, tag) {
        var cmpMeta = plt.getComponentMeta(tag);
        var initalLoad = false;
        var instance = elm.$instance;
        if (!instance) {
            instance = elm.$instance = new cmpMeta.componentModule();
            instance.$el = elm;
            instance.$destroys = [];
            instance.$onDestroy = function (cb) {
                instance.$destroys.push(cb);
            };
            initProps(plt, config, renderer, elm, tag, instance, cmpMeta.props, cmpMeta.watches);
            initalLoad = true;
        }
        if (!instance.$root) {
            var cmpMode = cmpMeta.modes[instance.mode];
            var cmpModeId = tag + '.' + instance.mode;
            instance.$root = plt.$attachShadow(elm, cmpMode, cmpModeId);
        }
        var vnode = generateVNode(instance.$root, instance, cmpMeta.hostCss);
        // if we already have a vnode then use it
        // otherwise, elm is the initial patch and
        // we need it to pass it the actual host element
        instance.$vnode = renderer(instance.$vnode ? instance.$vnode : elm, vnode);
        if (initalLoad && instance) {
            instance.ionViewDidLoad && instance.ionViewDidLoad();
        }
    }

    function connectedCallback(plt, config, renderer, elm, cmpMeta) {
        plt.nextTick(function () {
            var tag = cmpMeta.tag;
            var cmpMode = cmpMeta.modes[getMode(plt, config, elm, 'mode')];
            plt.loadComponent(cmpMode.bundleId, function loadComponentCallback() {
                queueUpdate(plt, config, renderer, elm, tag);
            });
        });
    }
    function getMode(plt, config, elm, propName) {
        var value = elm[propName];
        if (isDef(value)) {
            return value;
        }
        value = plt.$getAttribute(elm, propName);
        if (isDef(value)) {
            return value;
        }
        return config.get(propName, 'md');
    }

    function disconnectedCallback(elm) {
        if (elm) {
            var instance = elm.$instance;
            if (instance) {
                instance.ionViewWillUnload && instance.ionViewWillUnload();
                var destroys = instance.$destroys;
                if (destroys) {
                    for (var i = 0; i < destroys.length; i++) {
                        destroys[i]();
                    }
                    destroys.length = 0;
                }
                instance.$vnode && invokeDestroyHook(instance.$vnode);
                elm.$instance = instance.$el = instance.$root = instance.$vnode = null;
            }
        }
    }

    function registerComponents(renderer, plt, config, components) {
        Object.keys(components || {}).forEach(function (tag) {
            var cmpMeta = plt.registerComponent(tag, components[tag]);
            // closure doesn't support outputting es6 classes (yet)
            // build step does some closure tricks by changing this class
            // to just a function for compiling, then changing it back to a class
            class ProxyElement extends HTMLElement {}
            ProxyElement.prototype.connectedCallback = function () {
                connectedCallback(plt, config, renderer, this, cmpMeta);
            };
            ProxyElement.prototype.attributeChangedCallback = function (attrName, oldVal, newVal) {
                attributeChangedCallback(this, cmpMeta, attrName, oldVal, newVal);
            };
            ProxyElement.prototype.disconnectedCallback = function () {
                disconnectedCallback(this);
            };
            ProxyElement.observedAttributes = cmpMeta.obsAttrs;
            window.customElements.define(tag, ProxyElement);
        });
    }

    var ionic = window.Ionic = window.Ionic || {};
    // most of the controllers are added to window.Ionic within ionic-angular
    var plt = PlatformClient(window, document, ionic, ionic.staticDir, ionic.domCtrl, ionic.nextTickCtrl);
    var renderer = Renderer(plt);
    registerComponents(renderer, plt, ionic.configCtrl, ionic.components);
})(window, document);